/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __PIKA_PALETTE_LOAD_H__
#define __PIKA_PALETTE_LOAD_H__


#define PIKA_PALETTE_FILE_EXTENSION ".gpl"


typedef enum
{
  PIKA_PALETTE_FILE_FORMAT_UNKNOWN,
  PIKA_PALETTE_FILE_FORMAT_GPL,      /* PIKA palette                        */
  PIKA_PALETTE_FILE_FORMAT_RIFF_PAL, /* RIFF palette                        */
  PIKA_PALETTE_FILE_FORMAT_ACT,      /* Photoshop binary color palette      */
  PIKA_PALETTE_FILE_FORMAT_PSP_PAL,  /* JASC's Paint Shop Pro color palette */
  PIKA_PALETTE_FILE_FORMAT_ACO,      /* Photoshop ACO color file            */
  PIKA_PALETTE_FILE_FORMAT_ACB,      /* Photoshop ACB color book            */
  PIKA_PALETTE_FILE_FORMAT_ASE,      /* Photoshop ASE color palette         */
  PIKA_PALETTE_FILE_FORMAT_CSS,      /* Cascaded Stylesheet file (CSS)      */
  PIKA_PALETTE_FILE_FORMAT_SBZ       /* Swatchbooker SBZ file               */
} PikaPaletteFileFormat;


GList               * pika_palette_load               (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_act           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_riff          (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_psp           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_aco           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_acb           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_ase           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_css           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);
GList               * pika_palette_load_sbz           (PikaContext   *context,
                                                       GFile         *file,
                                                       GInputStream  *input,
                                                       GError       **error);

PikaPaletteFileFormat pika_palette_load_detect_format (GFile         *file,
                                                       GInputStream  *input);


#endif /* __PIKA_PALETTE_H__ */
